/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//---------------------------------------------------------------------------
#ifndef ScenixH
#define ScenixH
#include "project.h"


#define WDTPERIOD (1/14000)  //14kHz
#define SX_18 1
#define SX_20 2
#define SX_28 3

#define RST_Powerup   0x01
#define RST_Wakeup    0x02
#define RST_Brownout  0x04
#define RST_Timeout   0x08
#define RST_MCLR      0x10

class Sx1828;

class SxDisassembly : public disassembly
{
public:
  Sx1828* Sx;
  void ToggleBreak(int Addr);
  int IsBreak(int Addr);
  char *GetLine(int Addr);
  char *GetSourceCode(int Addr);
  int GetIP();
  void SetIP(int Addr);
  virtual int GetAddr();
  virtual void SetAddr(int Addr);
  void GetMemRange(int *start, int *end);
  char *GetOpCode(int Addr);
  int GetOpCodeLength(int Addr) { return 1; };
  void ToCursor(int Addr);
  void ClearBreakpoints();
  char *GetRealAddr(int Addr);
  virtual void ReloadCode();
  SxDisassembly(int name, attribute *attribs, symbol *parent);
};

class SxSource : public disassembly
{
public:
  Sx1828* Sx;
  void ToggleBreak(int Addr);
  int IsBreak(int Addr);
  char *GetLine(int Addr);
  char *GetSourceCode(int Addr);
  int GetIP();
  void SetIP(int Addr);
  virtual int GetAddr();
  virtual void SetAddr(int Addr);
  void GetMemRange(int *start, int *end);
  char *GetOpCode(int Addr);
  int GetOpCodeLength(int Addr) { return 1; };
  void ToCursor(int Addr);
  void ClearBreakpoints();
  char *GetRealAddr(int Addr);
  virtual void ReloadCode();
  SxSource(int name, attribute *attribs, symbol *parent);
};

class Sx1828type : public componenttype
{
public:
  component *create(int name, attribute *attribs, symbol *parent);
};

class Sx1828 : public component
{

public:
  word Fuse,Fusex;
  //Broches E/S
  pin_Input  OSC1,MCLR;
  pin_Edge   RTCC;
  pin_Output OSC2;
  pin_3state *PORT[3][8];
  //Variables internes voir doc SX
  byte SpecialPins; //Etat des broches spciales : b0: OSC1 b1:RTCC b2:MCLR b3:OSC2
                    //b4:Sortie OSC interne
  byte LastSpecialPins;
  byte LastPortB;
  byte R_OUT[3],LVL[3],PLP[3],TRIS[3];
  byte CMP_B,WKPND_B,WKED_B,WKEN_B,ST_B;
  byte ST_C;
  byte OPTION,MODE,W;
  byte WDT,WDT_PRE;    // Watchdog counter and prescaler
  word Stack[8];       // pile d'appels
  byte StackDepth;
  byte SP;             // Stack pointer
  word RTCCCounter;    // Compteur pour RTCC
  word WDTCounter;     // Compteur Watch Dog Timer
  word Prescaler;
  bool PowerDown;      // 1 indique SX en mode SLEEP
  int ToCursorBreakAddr;
  //Registres de sauvegarde d'interruption
  struct intstack {
    byte W,FSR,STATUS;
    word PC;
    int IntActive;
  } IntStack;
  //Memory
  word *ROM;
  byte *RAM;
  char** Lines; // Lines[addr] is source for address addr
  char** SourceLines; // SourceLine[n] is nth line of lst file.
  int NumLines;
  int *SourceToAddr;
  int *AddrToSource;
  
  void* *BreakPoints;
  word RomSize,RamSize;
  word PC;             //Program Counter
  int  CycleCount;
  int  SystemClock;
  real IClkPeriod;    //1/2 Priode de l'oscillateur interne
  real UpdateTime1;   //oscillateur interne
  real UpdateTime2;   //oscilalteur WDT
  //Methodes
  //  pin *GetPin(symbolname *name);
  void Update(int force = 0);
  Sx1828(componenttype *type, int name, attribute *attribs, symbol *parent,int SxType);
  void Reset(int type);  //Met le SX dans l'etat post-reset;
  virtual void _Reset() { Reset(RST_Powerup); }
  char * Disasm(word Addr);
  int Execute(void);
  void TestBreakPoints();
  int LoadHexFile(char* FileName);
  int LoadLstFile(char* FileName);
  void ReloadFiles();
  virtual ~Sx1828(void);
private:
  int StepStackPtr;
  int LastStepID;
  unsigned char Getfr(int n);
  void Setfr(int n,unsigned char v);
  void PushStack(word aPC);
  word PopStack(void);
  void ModifyPortConfig(byte reg);
  symbol *_getlocalsymbol(symbolname *);
  list<symbol> *getapparentsymbols();
};

//---------------------------------------------------------------------------
#endif
