/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __BUS_H__
#define __BUS_H__

class plugconstructor;

class plugtype
{
  list<symbolname> *symbolnames;
  int name;
  int numsymbols;
 public:
  int GetName() { return name; }
  int NumSymbols() { return numsymbols; }
  list<symbolname> *GetSymbolNames() { return symbolnames; }
  plugtype(int name, list<symbolname> *symbolnames);
};

class plug : public symbol
{
  friend class bus;
  friend class plugwidget;
protected:
  plugtype *type;
  list<symbol> *symbols;
  plug(int name, attribute *attributes, symbol *parent);
  symbol *_getlocalsymbol(symbolname *s);
public:    
  virtual list<symbol> *getapparentsymbols() { return symbols; }
  symbolwidget *getwidget();
  plug(int name, attribute *attributes, symbol *parent, plugconstructor *constr);
  virtual plug *ToPlug() { return this; }
  plugtype *gettype() { return type; }
};

class plugconstructor : public symbolconstructor
{
  friend plug;
  list<symbolname> *symbols;
protected:
  symbol *_create(symbol *parent);
  plugtype *type;
public:
  plugconstructor(int name, attribute *attribs, list<symbolname> *symbols, plugtype *type = (plugtype *) NULL);
};

class busconstructor : public symbolconstructor
{
  friend class bus;
  list <symbolname> *plugnames;
  int mixed : 1;
protected:
  symbol *_ccreate(macrocomponent *parent);
public:
  busconstructor(int name, attribute *attribs, list<symbolname> *plugnamesr, int _mixed = 0);
};

class bus : public plug
{
  list<plug> *plugs;
public:
  bus(int name, attribute *attributes, macrocomponent *parent, busconstructor *const);
};

#endif
