/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __COMPONENT_H__
#define __COMPONENT_H__

class component : public symbol
{
protected:
  componenttype *type;

  symbolwidget *getwidget() { return new projectwidget(getattributes(), this); }

  unsigned int updateneeded : 1;
  real updatetime;
  virtual void _Reset() {};

public:
  componenttype *GetType() { return type; }
  void SetUpdateNeeded();
  int UpdateNeeded() { return updateneeded; }
  real UpdateTime() {return updatetime; }
  virtual void Update(int force = 0) = 0;
  void Reset();
  virtual component *ToComponent() { return this; }

  component(componenttype *type, int name, attribute *attribs, symbol *parent);
};

typedef list<component> componentlist;

class macrocomponent : public component
{
  friend symbolconstructor;
  friend net;
  friend class netconstructor;
  friend componentconstructor;
  friend class macrocomponentconstructor;
  friend class macrocomponenttype;
  componentlist *components;
  netlist *nets;

  virtual symbol *_getlocalsymbol(symbolname *s);

public:
  virtual list<symbol> *getapparentsymbols();
  virtual void Update(int force = 0);

  virtual void _Reset();

  macrocomponent(componenttype *type, int nameid, attribute *attribs, symbol *parent);
};

class componentconstructor : public symbolconstructor
{
  macrocomponenttype *parent;
  symbol * _ccreate(macrocomponent *parent);
public:
  macrocomponenttype *getparent() { return parent; }
  componenttype *type;
  componentconstructor(int name, attribute *attribs, macrocomponenttype *parent);
};

#endif


