/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "project.h"

#define TARGET_MOVE 1
#define TARGET_HBOX 2
#define TARGET_VBOX 3
#define TARGET_PROJECT 4
#define TARGET_SYMBOL 5

#define NEW_PROJECT 1000

int idletype = IDLE_NONE;			
gint idletag;					
gint idledelay;

struct GtkTargetEntry WidgetTarget = { "symbolwidget", GTK_TARGET_SAME_APP, TARGET_MOVE}; 
struct GtkTargetEntry NewHBoxTarget = { "newhboxwidget", GTK_TARGET_SAME_APP, TARGET_HBOX}; 
struct GtkTargetEntry NewVBoxTarget = { "newvboxwidget", GTK_TARGET_SAME_APP, TARGET_VBOX}; 
struct GtkTargetEntry NewProjectTarget = { "newvprojectwidget", GTK_TARGET_SAME_APP, TARGET_PROJECT}; 
struct GtkTargetEntry NewSymbolTarget = { "newvsymbolwidget", GTK_TARGET_SAME_APP, TARGET_SYMBOL}; 
struct GtkTargetEntry AllTargets[] = { WidgetTarget, NewHBoxTarget, NewVBoxTarget, NewProjectTarget, NewSymbolTarget };

GtkStyle *InsensitiveStyle;
GtkStyle *NormalStyle;

void PrepareStyles()
{
  NormalStyle = gtk_style_new();
  InsensitiveStyle = gtk_style_new();

  InsensitiveStyle->fg[GTK_STATE_NORMAL] = InsensitiveStyle->fg[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->bg[GTK_STATE_NORMAL] = InsensitiveStyle->bg[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->light[GTK_STATE_NORMAL] = InsensitiveStyle->light[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->dark[GTK_STATE_NORMAL] = InsensitiveStyle->dark[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->mid[GTK_STATE_NORMAL] = InsensitiveStyle->mid[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->text[GTK_STATE_NORMAL] = InsensitiveStyle->text[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->base[GTK_STATE_NORMAL] = InsensitiveStyle->base[GTK_STATE_INSENSITIVE];

  InsensitiveStyle->fg[GTK_STATE_ACTIVE] = InsensitiveStyle->fg[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->bg[GTK_STATE_ACTIVE] = InsensitiveStyle->bg[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->light[GTK_STATE_ACTIVE] = InsensitiveStyle->light[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->dark[GTK_STATE_ACTIVE] = InsensitiveStyle->dark[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->mid[GTK_STATE_ACTIVE] = InsensitiveStyle->mid[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->text[GTK_STATE_ACTIVE] = InsensitiveStyle->text[GTK_STATE_INSENSITIVE];
  InsensitiveStyle->base[GTK_STATE_ACTIVE] = InsensitiveStyle->base[GTK_STATE_INSENSITIVE];
}

/****************************************************************************

			 IDLE FUNCTION HANDLING

 ****************************************************************************/

int setidleidle(int keep = 0)
{
  switch(idletype)
  {
    case IDLE_IDLE:
      return TRUE;
      break;
    case IDLE_DELAY:
      if (!keep) gtk_timeout_remove(idletag);
      break;
  }

  idletag = gtk_idle_add(Simulate, NULL);
  idletype = IDLE_IDLE;
  return FALSE;
}

int setidlenone(int keep = 0)
{
  switch(idletype)
  {
    case IDLE_DELAY:
      if (!keep) gtk_timeout_remove(idletag);
      break;
    case IDLE_IDLE:
      if (!keep) gtk_idle_remove(idletag);
      break;
  }

  idletype = IDLE_NONE;

  return FALSE;
}

int setidletimeout(gint time, int keep)
{
  if (time == 0) return setidleidle();
  if (time < 0) return setidlenone();

  switch(idletype)
  {
    case IDLE_DELAY:
      if (time == idledelay) return TRUE;
      if (!keep) gtk_timeout_remove(idletag);
      break; 
    case IDLE_IDLE:
      if (!keep) gtk_idle_remove(idletag);
      break;
  }

  idletag = gtk_timeout_add(time, Simulate, NULL);
  idletype = IDLE_DELAY;
  idledelay = time;
  return FALSE;
}

/****************************************************************************

			       BUTTON WITH VALUE

 ****************************************************************************/

void LabelWithValueChanged(GtkAdjustment *adjustment, int *value)
{
  *value = (int) adjustment->value;

  /*  if (Project.RunMode == RUN_SLOWANIM)
  {
    gtk_timeout_remove(idletag);
    idletag = gtk_timeout_add(Project.SlowAnimateDelay, Simulate, NULL);
    idletype = IDLE_DELAY;
    }*/
}

GtkWidget *LabelWithValue(char *Label, int *value, int digits, int step, int bigstep)
{
  GtkWidget *Edt, *Box, *Lbl;

  GtkAdjustment *adj = (GtkAdjustment *) gtk_adjustment_new(*value, 1, 1 << (8 * sizeof(int) - 2), step, bigstep, 0);
  gtk_signal_connect(GTK_OBJECT(adj), "value-changed", GTK_SIGNAL_FUNC(LabelWithValueChanged), value);

  Edt = gtk_spin_button_new (adj, 2, 0);
    GdkFont *fnt = gtk_widget_get_style(Edt)->font; // Kludge
    int ascent, descent, width;
    gdk_string_extents(fnt, "0", (int *) NULL, (int *) NULL, &width, &ascent, &descent);
    gtk_widget_set_usize(Edt, width * (digits + 2) + 8, ascent + descent + 8);
  gtk_widget_show(Edt);

  /*  Bord = gtk_frame_new(NULL);
  gtk_container_add(GTK_CONTAINER(Bord), Edt);
  gtk_widget_show(Bord);*/

  Lbl = gtk_label_new(Label);
  gtk_widget_show(Lbl);

  Box = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start_defaults(GTK_BOX(Box), Lbl);
  gtk_box_pack_start_defaults(GTK_BOX(Box), Edt);
  gtk_widget_show(Box);

  return Box;
}

/****************************************************************************

			       WIDGET

 ****************************************************************************/

void widget::_parseattribs(attribute *attribs)
{ 
}

void widget::localparseattribs(attribute *attribs)
{
  if (startup) attribbacklog = new list<attribute>(attribs, attribbacklog);
  else
  {
    _parseattribs(attribs);
    if (attribbacklog != NULL)
    {
      revlist(attribbacklog);
      for (list<attribute> *l = attribbacklog; l != NULL; l = l->next)
      {
        _parseattribs(l->elem);
        passtochildren((void (widget::*)(void *)) &widget::parseattribs, (void *) l->elem); 
      }

      delete attribbacklog;
      attribbacklog = (list<attribute> *) NULL;
    }
  }
}

void widget::parseattribs(attribute *attribs) 
{ 
  localparseattribs(attribs); 
  passtochildren((void (widget::*)(void *)) &widget::parseattribs, (void *) attribs); 
};

widget::~widget()
{
  // #### Err!!
}

void widget::display()
{
  if (startup) 
  {
    startup = 0;
    localparseattribs(initattribs);
  }
  draw();
  passtochildren(&symbolwidget::display);
  gtk_widget_show(wi);
}

void widget::refresh()
{
  draw();
  passtochildren(&symbolwidget::refresh);
}

widget::widget(attribute *attribs) 
{ 
  wi = (GtkWidget *) NULL; 
  initattribs = attribs;
  startup = 1;
  attribbacklog = (list<attribute> *) NULL;
}

void widget::draw()
{ 
}

void widget::passtochildren(void (widget::*fn)(void *), void *p)
{
}

void widget::RemoveFromContainer(GtkWidget *c)
{ 
  gtk_container_remove(GTK_CONTAINER(c), wi); 
}

/****************************************************************************

				 WINDOW

 ****************************************************************************/

void window::RemoveChild(symbolborderwidget *sbw)
{
  if (sbw != w) 
    error(TYPE_INTERNALERROR, -1, NULL, "Internal error : window::RemoveChild on non parent window.");

  sbw->RemoveFromContainer(this->box);

  w = (symbolborderwidget *) NULL;
  if (!destroying) delete this;
}

window::~window()
{
  if (destroying)
    error(TYPE_INTERNALERROR, -1, NULL, "Internal error : window::~window reentered.");

  destroying = 1; // Avoid recursive calls to the destructor
  //  gtk_signal_disconnect_by_func(GTK_OBJECT(wi), GTK_SIGNAL_FUNC(window::delete_event), this); // Avoid having the distructor called when we delete the GTK window...

  Project.UnlinkWindow(this);
  delete w; 
  gtk_widget_hide(wi);
  gtk_widget_destroy(wi);
}

void window::passtochildren(void (widget::*fn)(void *), void *p)
{
  (w->*fn)(p); 
}

void window::delete_event(GtkWidget *widget, GdkEvent  *event, window *_this)
{
  if (!_this->destroying) delete _this;
}

void window::newwindow(GtkButton *button, int type)
{
  symbol *s;
  if (type == NEW_PROJECT) s = &Project;
  else s = displaybox::getempty(type);
  window *newwin = new window(Project.getattributes(), s->getwidget());
  newwin->display();
}

void window::trashusedropdata(GtkWidget *widget, GdkDragContext *drag_context, gint x, gint y, GtkSelectionData *data, guint info, guint time,  gpointer udata)
{
  symbolborderwidget *symb = *(symbolborderwidget **) data->data;
  gtk_drag_finish(drag_context, TRUE, TRUE, time);
  
  delete symb;
}

void window::droptoeditzone(GtkWidget *widget, GdkDragContext *drag_context, gint x, gint y, GtkSelectionData *data, guint info, guint time,  gpointer udata)
{
  symbolborderwidget *symb = *(symbolborderwidget **) data->data;
  gtk_drag_finish(drag_context, TRUE, TRUE, time);

  gtk_entry_set_text(GTK_ENTRY(widget), symb->getsymbol()->getfullnametext());
}

void window::newwindowfromselection(GtkWidget *widget, GdkDragContext *drag_context, gint x, gint y, GtkSelectionData *data, guint info, guint time,  gpointer udata)
{
  window *newwin = new window(Project.getattributes(), *(symbolborderwidget **) data->data);
  newwin->display();
}

void window::getboxdata(GtkWidget *widget, GdkDragContext *drag_context, GtkSelectionData *data, guint info, guint time, window *_this)
{
  if (info == TARGET_SYMBOL)
  {
    char *name = gtk_entry_get_text(GTK_ENTRY(widget));
    gtk_selection_data_set(data, 0, 8, (guchar *) name, strlen(name) + 1); 
  }
  else gtk_selection_data_set(data, 0, 0, (guchar *) NULL, 0);
}

void window::editzonehighlight(GtkWidget *wid)
{
  static int recurs = 0;
  
  if (recurs) return;

  recurs = 1;

  if (Project.getsymbol(gtk_entry_get_text(GTK_ENTRY(wid))) != NULL)
    gtk_widget_set_style(wid, NormalStyle);
  else
    gtk_widget_set_style(wid, InsensitiveStyle);

  recurs = 0;
}

void window::completeeditzone(GtkWidget *wid)
{
  char *prevtxt = strdup(gtk_entry_get_text(GTK_ENTRY(wid)));
  gtk_editable_delete_selection(GTK_EDITABLE(wid));
  char *basetxt = strdup(gtk_entry_get_text(GTK_ENTRY(wid)));

  int findfirst = 1;
  if (strncmp(prevtxt, basetxt, strlen(basetxt)) == 0 && strcmp(prevtxt, basetxt) != 0) findfirst = 0;

  //  cout << prevtxt << " " << basetxt << " " << findfirst << "\n";

  int slashpos;
  for (slashpos = strlen(basetxt) - 1; slashpos >= 0 && basetxt[slashpos] != '/'; slashpos--);

  symbol *basesymb;
  if (slashpos < 0) basesymb = &Project;
  else 
  {
    basetxt[slashpos] = 0;
    basesymb = Project.getsymbol(basetxt);
    basetxt[slashpos] = '/';
    if (basesymb == NULL)
    {
      free(prevtxt);
      free(basetxt);
      return;
    }
  }

  int namestart;
  for (namestart = slashpos + 1; basetxt[namestart] == ' ' || basetxt[namestart] == '\t'; namestart++);

  int nameend;
  for (nameend = strlen(basetxt) - 1; nameend >= 0 && (basetxt[nameend] == ' ' || basetxt[nameend] == '\t'); nameend--);

  if (nameend < 0)
  {
      free(prevtxt);
      free(basetxt);
      return;
  }

  for (list<symbol> *l = basesymb->getapparentsymbols(); l != NULL; l = l->next)
  {
    if (symbtable.isanonymous(l->elem->getname())) continue;
    char *curname = symbtable.tostring(l->elem->getname());
    if (strncmp(curname, basetxt + namestart, nameend - namestart + 1) == 0)
    {
      if (findfirst)
      {
	int newend = nameend + 1;
        gtk_editable_insert_text(GTK_EDITABLE(wid), curname + (nameend - namestart + 1), strlen(curname) - (nameend - namestart + 1) + namestart, &newend);
        gtk_entry_set_position(GTK_ENTRY(wid), nameend + 1);
	gtk_editable_select_region(GTK_EDITABLE(wid), nameend + 1, -1);//namestart + strlen(curname));
        free(prevtxt);
        free(basetxt);
        return;
      }
      else if (strcmp(curname, prevtxt + namestart) == 0)
      {
	findfirst = 1;
      }
    }
  }
  free(prevtxt);
  free(basetxt);
  return;
}

window::window(attribute *attribs, symbolwidget *sw) : widget(attribs)
{
  construct(new symbolborderwidget(attribs, sw));
}

window::window(attribute *attribs, symbolborderwidget *sw) : widget(attribs)
{
  construct(sw);
}

void window::setrunmode(GtkButton *button, int mode) 
{ 
  Project.RunMode = mode; 

  //  cout << "SetRunMode " << hex << mode << "\n";

  if (Project.RunMode == RUN_GO) Project.RedrawAll(); 
  
  setidleidle();
}

void window::exportsig(GtkButton *button, window *_this)
{
  _this->export(0);
}

void window::reset()
{
  Project.Reset();
}

void window::construct(symbolborderwidget *sw)
{
  OldRunMode = -1;
  destroying = 0;
  GtkWidget *CurButton, *CurBox;

  // Toolbar preparation

  editbox = gtk_hbox_new(FALSE, 0);
  runbox = gtk_hbox_new(FALSE, 0);

  // Edit Buttons
  CurButton = gtk_button_new_with_label("Run\nMode");
  gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::runmodeclicked), this);
  gtk_widget_show(CurButton);
  gtk_box_pack_start(GTK_BOX(editbox), CurButton, FALSE, TRUE, 0);

  CurButton = gtk_button_new_with_label("New Window\nFrom Selection");
  gtk_drag_dest_set(CurButton, GTK_DEST_DEFAULT_ALL, &WidgetTarget, 1, (enum GdkDragAction) (GDK_ACTION_MOVE | GDK_ACTION_COPY));
  gtk_signal_connect(GTK_OBJECT(CurButton), "drag-data-received", GTK_SIGNAL_FUNC(window::newwindowfromselection), this);
  gtk_widget_show(CurButton);
  gtk_box_pack_start(GTK_BOX(editbox), CurButton, FALSE, TRUE, 0);

  CurButton = gtk_button_new_with_label("New\nHorizontal");
  gtk_drag_source_set(CurButton, GDK_BUTTON2_MASK, &NewHBoxTarget, 1, GDK_ACTION_MOVE);
  gtk_signal_connect(GTK_OBJECT (CurButton), "drag-data-get", GTK_SIGNAL_FUNC(window::getboxdata), this);
  gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::newwindow), (gpointer) BOX_H);
  gtk_widget_show(CurButton);
  gtk_box_pack_start(GTK_BOX(editbox), CurButton, FALSE, TRUE, 0);

  CurButton = gtk_button_new_with_label("New\nVertical");
  gtk_drag_source_set(CurButton, GDK_BUTTON2_MASK, &NewVBoxTarget, 1, GDK_ACTION_MOVE);
  gtk_signal_connect(GTK_OBJECT (CurButton), "drag-data-get", GTK_SIGNAL_FUNC(window::getboxdata), this);
  gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::newwindow), (gpointer) BOX_V);
  gtk_widget_show(CurButton);
  gtk_box_pack_start(GTK_BOX(editbox), CurButton, FALSE, TRUE, 0);

  CurButton = gtk_button_new_with_label("New\nProject");
  gtk_drag_source_set(CurButton, GDK_BUTTON2_MASK, &NewProjectTarget, 1, GDK_ACTION_MOVE);
  gtk_signal_connect(GTK_OBJECT (CurButton), "drag-data-get", GTK_SIGNAL_FUNC(window::getboxdata), this);
  gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::newwindow), (gpointer) NEW_PROJECT);
  gtk_widget_show(CurButton);
  gtk_box_pack_start(GTK_BOX(editbox), CurButton, FALSE, TRUE, 0);

  CurButton = gtk_button_new_with_label("Export");
  gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::exportsig), this);
  gtk_widget_show(CurButton);
  gtk_box_pack_start(GTK_BOX(editbox), CurButton, FALSE, TRUE, 0);

  CurButton = gtk_button_new_with_label("Trash\nCan");
  gtk_drag_dest_set(CurButton, GTK_DEST_DEFAULT_ALL, &WidgetTarget, 1, GDK_ACTION_MOVE);
  gtk_signal_connect(GTK_OBJECT(CurButton), "drag-data-received", GTK_SIGNAL_FUNC(window::trashusedropdata), this);
  gtk_widget_show(CurButton);
  gtk_box_pack_start(GTK_BOX(editbox), CurButton, FALSE, TRUE, 0);
  gtk_signal_connect(GTK_OBJECT (CurButton), "drag-data-get", GTK_SIGNAL_FUNC(window::getboxdata), this);

  CurBox = gtk_vbox_new(FALSE, 0);

    CurButton = gtk_label_new("Insert symbol :");
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_entry_new();
    gtk_drag_source_set(CurButton, (enum GdkModifierType) (GDK_BUTTON2_MASK | GDK_BUTTON1_MASK), &NewSymbolTarget, 1, GDK_ACTION_MOVE);
    gtk_signal_connect(GTK_OBJECT (CurButton), "drag-data-get", GTK_SIGNAL_FUNC(window::getboxdata), this);
    gtk_drag_dest_set(CurButton, GTK_DEST_DEFAULT_ALL, &WidgetTarget, 1, GDK_ACTION_COPY);
    gtk_signal_connect(GTK_OBJECT(CurButton), "drag-data-received", GTK_SIGNAL_FUNC(window::droptoeditzone), this);
    gtk_signal_connect(GTK_OBJECT(CurButton), "changed", GTK_SIGNAL_FUNC(window::editzonehighlight), (gpointer) NULL);
    gtk_signal_connect(GTK_OBJECT(CurButton), "activate", GTK_SIGNAL_FUNC(window::completeeditzone), (gpointer) 0);
    gtk_widget_show(CurButton);
    gtk_widget_set_style(CurButton, InsensitiveStyle);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

  gtk_box_pack_start(GTK_BOX(editbox), CurBox, FALSE, TRUE, 0);
  gtk_widget_show(CurBox);

  // Run buttons
 
  CurBox = gtk_vbox_new(FALSE, 0);

    CurButton = gtk_button_new_with_label("Edit Mode");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::editmodeclicked), this);
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_button_new_with_label("Reset");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::reset), NULL);
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_button_new_with_label("Stop");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) RUN_STOP);
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

  gtk_widget_show(CurBox);
  gtk_box_pack_start(GTK_BOX(runbox), CurBox, FALSE, TRUE, 0);
  CurBox = gtk_vbox_new(FALSE, 0);

    CurButton = gtk_button_new_with_label("Go...");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) RUN_GO);
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_button_new_with_label("Fast");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) (RUN_ANIM | RUN_FASTANIM | RUN_TICK));
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_button_new_with_label("To Cursor");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) RUN_GO);
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

  gtk_widget_show(CurBox);
  gtk_box_pack_start(GTK_BOX(runbox), CurBox, FALSE, TRUE, 0);

  CurButton = gtk_label_new("Animate : ");
  gtk_widget_show(CurButton);
  gtk_box_pack_start(GTK_BOX(runbox), CurButton, FALSE, TRUE, 0);

  CurBox = gtk_vbox_new(FALSE, 0);

    CurButton = gtk_button_new_with_label("Tick");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) (RUN_ANIM | RUN_NORMANIM | RUN_TICK));
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_button_new_with_label("Trace");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) (RUN_ANIM | RUN_NORMANIM | RUN_TRACE));
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_button_new_with_label("Step");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) (RUN_ANIM | RUN_NORMANIM | RUN_STEP));
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

  gtk_widget_show(CurBox);
  gtk_box_pack_start(GTK_BOX(runbox), CurBox, FALSE, TRUE, 0);

  CurButton = gtk_label_new("Slow : ");
  gtk_widget_show(CurButton);
  gtk_box_pack_start(GTK_BOX(runbox), CurButton, FALSE, TRUE, 0);

  CurBox = gtk_vbox_new(FALSE, 0);

    CurButton = gtk_button_new_with_label("Tick");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) (RUN_ANIM | RUN_SLOWANIM | RUN_TICK));
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_button_new_with_label("Trace");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) (RUN_ANIM | RUN_SLOWANIM | RUN_TRACE));
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_button_new_with_label("Step");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) (RUN_ANIM | RUN_SLOWANIM | RUN_STEP));
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

  gtk_widget_show(CurBox);
  gtk_box_pack_start(GTK_BOX(runbox), CurBox, FALSE, TRUE, 0);

  CurButton = gtk_label_new("One : ");
  gtk_widget_show(CurButton);
  gtk_box_pack_start(GTK_BOX(runbox), CurButton, FALSE, TRUE, 0);

  CurBox = gtk_vbox_new(FALSE, 0);

    CurButton = gtk_button_new_with_label("Tick");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) (RUN_ANIM | RUN_ONCE | RUN_TICK));
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_button_new_with_label("Trace");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) (RUN_ANIM | RUN_ONCE | RUN_TRACE));
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

    CurButton = gtk_button_new_with_label("Step");
    gtk_signal_connect(GTK_OBJECT(CurButton), "clicked", GTK_SIGNAL_FUNC(window::setrunmode), (gpointer) (RUN_ANIM | RUN_ONCE | RUN_STEP));
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, FALSE, TRUE, 0);

  gtk_widget_show(CurBox);
  gtk_box_pack_start(GTK_BOX(runbox), CurBox, FALSE, TRUE, 0);
  CurBox = gtk_vbox_new(FALSE, 0);

    CurButton = LabelWithValue(" Fast speed ", &Project.FastAnimateSpeed, 5, 10, 500);
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, TRUE, TRUE, 0);

    CurButton = LabelWithValue(" Slow delay ", &Project.SlowAnimateDelay, 5, 1, 50);
    gtk_widget_show(CurButton);
    gtk_box_pack_start(GTK_BOX(CurBox), CurButton, TRUE, TRUE, 0);

  gtk_widget_show(CurBox);
  gtk_box_pack_start(GTK_BOX(runbox), CurBox, FALSE, TRUE, 0);

  // Tool bar show

  gtk_widget_show(runbox);

  // Main window contents
  window::w = sw;

  box = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box), editbox, FALSE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(box), runbox, FALSE, TRUE, 0);
  gtk_box_pack_start_defaults(GTK_BOX(box), w->GrabGtkWidget(this, (GtkWidget *) NULL));
  gtk_widget_show(box);

  // Main window

  wi = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect(GTK_OBJECT (wi), "delete_event", GTK_SIGNAL_FUNC(window::delete_event), this);
  gtk_container_add(GTK_CONTAINER(wi), box);
  gtk_window_set_policy(GTK_WINDOW(wi), FALSE, TRUE, TRUE);
  gtk_window_add_accel_group(GTK_WINDOW(wi), symbolborderwidget::GetPopupMenuAccelGroup());

  Project.windows = new list<window>(this, Project.windows);
  
  runmode();
  display();
}

void window::_runmode()
{
  gtk_widget_hide(editbox);
  gtk_widget_show(runbox);
}

void window::_editmode()
{
  gtk_widget_hide(runbox);
  gtk_widget_show(editbox);
}

void window::draw()
{
  char buff[100];


  if (Project.RunMode != OldRunMode)
  {
    OldRunMode = Project.RunMode;
    snprintf(buff, 100, "%s - %s", ProgramVersion, OldRunMode ? "Running" : "Stopped");
    gtk_window_set_title(GTK_WINDOW (wi), buff);
  }
}

void window::export(int indent)
{ 
  cout << "*****************************************\n\n";
  w->export(indent); 
  cout << "\n\n*****************************************\n";
}

/****************************************************************************

			SYMBOLBORDERWIDGET

 ****************************************************************************/

void symbolborderwidget::_parseattribs(attribute *attribs)
{
  int val;
  int fullnamepos;
  if ((fullnamepos = attribs->getintattribute("fullname", &val)))
  {
    displaymode.full = val;
    displaymode.title = 1;
    displaymode.extname = 0;
  }

  int titlepos;
  if ((titlepos = attribs->getintattribute("title", &val, fullnamepos)))
    displaymode.title = val;

  char *name;
  int disptitlepos;
  if ((disptitlepos = attribs->getstringattribute("disptitle", &name, fullnamepos)))
  {
    extname = symbtable.toid(name);
    if (!fullnamepos || fullnamepos > disptitlepos) displaymode.extname = 1;
  }

  if (fullnamepos || titlepos || disptitlepos) settitle();

  int hasbord;
  if (attribs->getintattribute("border", &hasbord) && (hasbord == 0) != (displaymode.hasborder == 0))
  {
    displaymode.hasborder = (hasbord != 0);
    GtkWidget *symbwid = s->GetGtkWidget(this);
    gtk_widget_ref(symbwid);
    if (!displaymode.hasborder)
    {
      gtk_container_remove(GTK_CONTAINER(border), symbwid);
      gtk_widget_show(title);
      gtk_widget_hide(border);
      gtk_box_pack_end(GTK_BOX(box), symbwid, canexpand(), TRUE, 0);
    }
    else
    {
      gtk_container_remove(GTK_CONTAINER(box), symbwid);
      gtk_container_add(GTK_CONTAINER(border), symbwid);
      gtk_widget_hide(title);
      gtk_widget_show(border);
    }
    settitle();
  }
}

void symbolborderwidget::exportattribs(int putcomma)
{
  if (putcomma) cout << ",";

  cout << "title = " << displaymode.title << ", border = " << displaymode.hasborder;
  if (displaymode.title)
  {
    if (displaymode.extname) cout << ", disptitle = \"" << symbtable.tostring(extname) << "\"";
    else cout << ", fullname = " << displaymode.full;
  }
}

void symbolborderwidget::export(int indent) 
{ 
  s->export(indent); 
}

symbol *symbolborderwidget::getsymbol()
{ 
  return s->getsymbol(); 
}

symbolborderwidget::~symbolborderwidget()
{
  if (Activated == this->wi) Activated = 0;
  GrabGtkWidget((widget *) NULL, (GtkWidget *) NULL);
  delete s;
}

void symbolborderwidget::passtochildren(void (widget::*fn)(void *), void *p)
{
  (s->*fn)(p);
}

GtkWidget *symbolborderwidget::Activated = (GtkWidget *) NULL;

void symbolborderwidget::activate(GtkWidget *widget, GdkEventCrossing *event, symbolborderwidget *_this)
{
  if (Activated != NULL) gtk_widget_set_state(Activated, GTK_STATE_NORMAL);
  gtk_widget_set_state(widget, GTK_STATE_ACTIVE);
  popupmenuwidget = _this;
  Activated = _this->wi;
}

/*void symbolborderwidget::disactivate(GtkWidget *widget, GdkEventCrossing *event, symbolborderwidget *_this)
{
  gtk_widget_set_state(widget, GTK_STATE_NORMAL);
  if (Activated == _this-> wi) Activated = (GtkWidget *) NULL;
}
*/

void symbolborderwidget::menusetattribute(gpointer callback_data, attribute *attribs, GtkWidget *widget)
{
  popupmenuwidget->parseattribs(attribs);
  Project.RedrawAll();
}

void symbolborderwidget::menusetattributelocal(gpointer callback_data, attribute *attribs, GtkWidget *widget)
{
  popupmenuwidget->_parseattribs(attribs);
  Project.RedrawAll();
}

typedef void (*ItemFactFunc)(...);

#define cNULL ((char *) NULL)
#define fNULL ((ItemFactFunc) NULL)

GtkItemFactoryEntry symbolborderwidget::PopupMenu[] = 
{
  { "/_Quit",         		"<alt>X",       (ItemFactFunc) gtk_main_quit, 0, cNULL },
  { "/_Base",         		cNULL,         	fNULL, 0, "<Branch>" },
  { "/Base/_Binary",  		"<control>B", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("base"), 2), cNULL },
  { "/Base/_Octal",    		"<control>O", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("base"), 8), cNULL },
  { "/Base/_Decimal", 		"<control>D", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("base"), 10), cNULL },
  { "/Base/_Hexadecimal", 	"<control>H", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("base"), 16), cNULL },
  { "/_Sign",      		cNULL,         	fNULL, 0, "<Branch>" },
  { "/Sign/_Unsigned",  	"<control>U", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("signed"), 0), cNULL },
  { "/Sign/Si_gned",  		"<control>G", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("signed"), 1), cNULL },
  { "/_Appearance",      		cNULL,         	fNULL, 0, "<Branch>" },
  { "/Appearance/_Full name",  		"<control>F", 	(ItemFactFunc) menusetattributelocal, (guint) intattribute(symbtable.toid("fullname"), 1), cNULL },
  { "/Appearance/_Short name",  		"<control>S", 	(ItemFactFunc) menusetattributelocal, (guint) intattribute(symbtable.toid("fullname"), 0), cNULL },
  { "/Appearance/_No name",  		"<control>N", 	(ItemFactFunc) menusetattributelocal, (guint) intattribute(symbtable.toid("title"), 0), cNULL },
  { "/Appearance/sep1", cNULL, 	fNULL, 0, "<Separator>" },
  { "/Appearance/With _Border", "<control>I", (ItemFactFunc) menusetattributelocal, (guint) intattribute(symbtable.toid("border"), 1), cNULL },
  { "/Appearance/_Without Border", "<control>T", (ItemFactFunc) menusetattributelocal, (guint) intattribute(symbtable.toid("border"), 0), cNULL },
  { "/Appearance/sep2", cNULL, 	fNULL, 0, "<Separator>" },
  { "/Appearance/_Grow", "KP_Add", (ItemFactFunc) menusetattributelocal, (guint) intattribute(symbtable.toid("border"), 1), cNULL },
  { "/Appearance/S_hrink", "KP_Subtract", (ItemFactFunc) menusetattributelocal, (guint) intattribute(symbtable.toid("border"), 0), cNULL },
  { "/_Data",      		cNULL,         	fNULL, 0, "<Branch>" },
  { "/Data/_Clear",  		"<control>Z", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("clear"), 1), cNULL },
  { "/Data/_Randomize",  	"<control>R", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("randomize"), 0), cNULL },
  { "/Data/_Increment",		"KP_Add", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addvalue"), 1), cNULL },
  { "/Data/_Decrement",		"KP_Subtract", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addvalue"), -1), cNULL },
  { "/Data/Add 10",		"<shift>KP_Add", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), 1), cNULL },
  { "/Data/Subtract 10",	"<shift>KP_Subtract", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), -1), cNULL },
  { "/Data/Add 100",		"<control>KP_Add", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), 2), cNULL },
  { "/Data/Subtract 100",	"<control>KP_Subtract", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), -2), cNULL },
  { "/Data/Add 1000",		"<control><shift>KP_Add", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), 3), cNULL },
  { "/Data/Subtract 1000",	"<control><shift>KP_Subtract", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), -3), cNULL },
  { "/Data/Add 10000",		"<alt>KP_Add", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), 4), cNULL },
  { "/Data/Subtract 10000",	"<alt>KP_Subtract", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), -4), cNULL },
  /*  { "/Data/Add 100000",		"<alt><shift>KP_Add", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), 5), cNULL },
  { "/Data/Subtract 100000",	"<alt><shift>KP_Subtract", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), -5), cNULL },
  { "/Data/Add 1000000",	"<alt><control>KP_Add", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), 6), cNULL },
  { "/Data/Subtract 1000000",	"<alt><control>KP_Subtract", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), -6), cNULL },
  { "/Data/Add 10000000",	"<alt><shift><control>KP_Add", 		(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), 7), cNULL },
  { "/Data/Subtract 10000000",	"<alt><shift><control>KP_Subtract",	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@addbasepow"), -7), cNULL },*/
  { "/_Array",      		cNULL,         	fNULL, 0, "<Branch>" },
  { "/Array/_1 column",  		"<control>1", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("columns"), 1), cNULL },
  { "/Array/_2 columns",  		"<control>2", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("columns"), 2), cNULL },
  { "/Array/_4 colmuns",  		"<control>4", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("columns"), 4), cNULL },
  { "/Array/_8 columns",  		"<control>8", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("columns"), 8), cNULL },
  { "/Array/_16 columns",  		"<control>6", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("columns"), 16), cNULL },
  { "/Array/sep1", cNULL, 	fNULL, 0, "<Separator>" },
  { "/Array/_Byte",  		"<control>Y", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("datawidth"), 8), cNULL },
  { "/Array/_Word",  		"<control>W", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("datawidth"), 16), cNULL },
  { "/_Run",      		cNULL,         	fNULL, 0, "<Branch>" },
  { "/Run/Set _Breakpoint",  	"F2", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@SetBreakPoint"), 1) },
  { "/Run/C_lear Breakpoints",  "F3", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@ClearBreakPoints"), 1) },
  { "/Run/To _Cursor",  	"F4", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@ToCursor"), 1) },
  { "/Run/_Stop",  		"F5", 	(ItemFactFunc) setrunmode, RUN_STOP, cNULL },
  { "/Run/_Tick",  		"F6", 	(ItemFactFunc) setrunmode, RUN_ANIM | RUN_ONCE | RUN_TICK, cNULL },
  { "/Run/_Trace",  		"F7", 	(ItemFactFunc) setrunmode, RUN_ANIM | RUN_ONCE | RUN_TRACE, cNULL },
  { "/Run/_Step",  		"F8", 	(ItemFactFunc) setrunmode, RUN_ANIM | RUN_ONCE | RUN_STEP, cNULL },
  { "/Run/_Go",  		"F9", 	(ItemFactFunc) setrunmode, RUN_GO, cNULL },
  { "/Run/Set _IP",  		"F10", 	(ItemFactFunc) menusetattribute, (guint) intattribute(symbtable.toid("@@SetIP"), 1) },
  { "/Run/Animate Tick Slow",	"<shift>F6", 	(ItemFactFunc) setrunmode, RUN_ANIM | RUN_SLOWANIM | RUN_TICK, cNULL },
  { "/Run/Animate Trace Slow",	"<shift>F7", 	(ItemFactFunc) setrunmode, RUN_ANIM | RUN_SLOWANIM | RUN_TRACE, cNULL },
  { "/Run/Animate Step Slow", "<shift>F8", 	(ItemFactFunc) setrunmode, RUN_ANIM | RUN_SLOWANIM | RUN_STEP, cNULL },
  { "/Run/Animate Tick Normal",	"<control>F6", 	(ItemFactFunc) setrunmode, RUN_ANIM | RUN_NORMANIM | RUN_TICK, cNULL },
  { "/Run/Animate Trace Normal","<control>F7", 	(ItemFactFunc) setrunmode, RUN_ANIM | RUN_NORMANIM | RUN_TRACE, cNULL },
  { "/Run/Animate Step Normal", "<control>F8", 	(ItemFactFunc) setrunmode, RUN_ANIM | RUN_NORMANIM | RUN_STEP, cNULL },
  { "/Run/Fast", "<control>F9", (ItemFactFunc) setrunmode, RUN_ANIM | RUN_FASTANIM | RUN_TRACE, cNULL },
};

void symbolborderwidget::setrunmode(gpointer callback_data, int mode, GtkWidget *widget)
{
  Project.RunMode = mode; 

  //  cout << "SetRunMode " << hex << mode << "\n";

  if (Project.RunMode == RUN_GO) Project.RedrawAll(); 
  
  setidleidle();
}

GtkItemFactory *symbolborderwidget::CreatePopupMenuFactory(GtkAccelGroup **accel)
{
  GtkItemFactory *PopupMenuFactory;
  GtkAccelGroup *accel_group;
  gint nmenu_items = sizeof(PopupMenu) / sizeof (PopupMenu[0]);
  accel_group = gtk_accel_group_new();
  PopupMenuFactory = gtk_item_factory_new(GTK_TYPE_MENU, "<main>", accel_group);
  gtk_item_factory_create_items(PopupMenuFactory, nmenu_items, PopupMenu, NULL);

  if (accel != NULL) *accel = accel_group;

  return PopupMenuFactory;
}

GtkAccelGroup *symbolborderwidget::GetPopupMenuAccelGroup()
{
  GtkAccelGroup *retval;
  CreatePopupMenuFactory(&retval);
  return retval;
}

GtkItemFactory *symbolborderwidget::GetPopupMenuFactory(GtkAccelGroup **accel)
{
  static GtkAccelGroup *accel_group;
  static GtkItemFactory *PopupMenuFactory = CreatePopupMenuFactory(&accel_group);

  if (accel != NULL) *accel = accel_group;

  return PopupMenuFactory;
}

symbolborderwidget *symbolborderwidget::popupmenuwidget = (symbolborderwidget *) NULL;

gboolean symbolborderwidget::popupmenu(GtkWidget *widget, GdkEventButton *event, symbolborderwidget *_this)
{
  if (event->button == 3 && event->type == GDK_BUTTON_PRESS)
  {
    gtk_item_factory_popup(GetPopupMenuFactory(), (int) event->x_root, (int) event->y_root, event->button, event->time);
    popupmenuwidget = _this;
    return TRUE;
  }
  
  return FALSE;
}

symbolborderwidget::symbolborderwidget(attribute *attribs, symbolwidget *s) : widget(attribs)
{
  for (list<attribute> *l = s->attribbacklog; l != NULL; l = l->next)
    attribbacklog = new list<attribute>(l->elem, attribbacklog);

  AssociatedSeparator = (GtkWidget *) NULL;
  symbolborderwidget::s = s;
  displaymode.title = !symbtable.isanonymous(s->getsymbol()->getname());
  displaymode.full = 0;
  displaymode.hasborder = s->hasborder();
  displaymode.extname = 0;
  parent = 0;

  box = gtk_hbox_new(FALSE, 0);
  gtk_box_set_spacing(GTK_BOX(box), 10);

  title = gtk_label_new("");
  gtk_box_pack_start(GTK_BOX(box), title, FALSE, TRUE, 0);

  border = gtk_frame_new("");
  gtk_container_set_border_width(GTK_CONTAINER(border), 5);
  gtk_box_pack_end(GTK_BOX(box), border, TRUE, TRUE, 0);

  if (!displaymode.hasborder)
  {
    gtk_widget_show(title);
    gtk_box_pack_end(GTK_BOX(box), s->GetGtkWidget(this), canexpand(), TRUE, 0);
  }
  else
  {
    gtk_container_add(GTK_CONTAINER(border), s->GetGtkWidget(this));
    gtk_widget_show(border);
  }
  settitle();
  gtk_widget_show(box);

  wi = gtk_event_box_new();
  gtk_container_add(GTK_CONTAINER(wi), box);
  gtk_drag_source_set(wi, GDK_BUTTON2_MASK, &WidgetTarget, 1, (enum GdkDragAction) (GDK_ACTION_MOVE | GDK_ACTION_COPY));
  gtk_signal_connect(GTK_OBJECT(wi), "drag-begin", GTK_SIGNAL_FUNC(begindrag), this);
  gtk_signal_connect(GTK_OBJECT(wi), "drag-end", GTK_SIGNAL_FUNC(enddrag), this);
  gtk_signal_connect(GTK_OBJECT(wi), "drag-data-get", GTK_SIGNAL_FUNC(getdropdata), this);
  gtk_signal_connect(GTK_OBJECT(wi), "enter-notify-event", GTK_SIGNAL_FUNC(activate),this);
  //  gtk_signal_connect(GTK_OBJECT(wi), "leave-notify-event", GTK_SIGNAL_FUNC(symbolborderwidget::disactivate), this);
  gtk_signal_connect(GTK_OBJECT(wi), "button-press-event", GTK_SIGNAL_FUNC(popupmenu), this);
}

void symbolborderwidget::symbolborderwidget::begindrag(GtkWidget *widget, GdkDragContext *drag_context, symbolborderwidget *_this)
{
  if (_this->parent->GetWindowWidget() != NULL) gtk_widget_hide(_this->parent->GetWindowWidget());
  gtk_widget_hide(_this->wi);
  if (_this->AssociatedSeparator != NULL) gtk_widget_hide(_this->AssociatedSeparator);
}

void symbolborderwidget::enddrag(GtkWidget *widget, GdkDragContext *drag_context, symbolborderwidget *_this)
{
  if ((drag_context->action & GDK_ACTION_MOVE) == 0)
  {
    gtk_widget_show(_this->wi);
    if (_this->AssociatedSeparator != NULL) gtk_widget_show(_this->AssociatedSeparator);
    if (_this->parent->GetWindowWidget() != NULL) gtk_widget_show(_this->parent->GetWindowWidget());
  }
}

void symbolborderwidget::getdropdata(GtkWidget *widget, GdkDragContext *drag_context, GtkSelectionData *data, guint info, guint time, symbolborderwidget *_this)
{
  gtk_selection_data_set(data, 0, sizeof(symbolborderwidget *) * 8, (guchar *) &_this, sizeof(symbolborderwidget *));
}

void symbolborderwidget::settitle()
{
  char *txt;
 
  if (!displaymode.extname)
  {
    if (!displaymode.title) txt = (char *) NULL;
    else txt = displaymode.full ? s->getsymbol()->getfullnametext() : symbtable.tostring(s->getsymbol()->getname());
  }
  else
  {
    txt = symbtable.tostring(extname);
  }
  
  if (displaymode.hasborder) gtk_frame_set_label(GTK_FRAME(border), txt);
  else gtk_label_set_text(GTK_LABEL(title), txt);
}

GtkWidget *symbolborderwidget::GrabGtkWidget(widget *p, GtkWidget *AssocSep) 
{ 
  if (parent != NULL) 
  {
    parent->RemoveChild(this);
  }

  AssociatedSeparator = AssocSep;
  parent = p; 
  return wi; 
}

/****************************************************************************

				   SYMBOLWIDGET

 ****************************************************************************/

symbolwidget::symbolwidget(attribute *attribs, symbol *s) : widget(attribs)
{
  symbolwidget::s = s;
}

void symbolwidget::export(int indent) 
{ 
  for (int i = 0; i < indent; i++) cout << " ";
  cout << "symbol " << "(";
  parent->exportattribs(exportattribs());
  cout << ") " << s->getfullnametext() << ";\n";
}


/****************************************************************************

				   PLAINSYMBOLWIDGET

 ****************************************************************************/

plainsymbolwidget::plainsymbolwidget(attribute *attribs, symbol *s) : symbolwidget(attribs, s)
{
  char buffer[100];
  snprintf(buffer, 100, "No widget for %s", s->getfullnametext());
  wi = gtk_label_new(buffer);
}

/****************************************************************************

				   PINWIDGET

 ****************************************************************************/

pinwidget::pinwidget(attribute *attribs, pin *p) : symbolwidget(attribs, p)
{
  wi = gtk_label_new("");
}

void pinwidget::draw()
{
  char buff[100];

  pin *p = (pin *) getsymbol();

  snprintf(buff, 100, "Vin(%.3f) Vout(%.3f) Z(%.3f)", p->GetVin(), p->GetVout(), p->GetZ());
  gtk_label_set_text(GTK_LABEL(wi), buff);
}

/****************************************************************************

				   BOXWIDGET

 ****************************************************************************/

void boxwidget::export(int indent)
{
  for (int i = 0; i < indent; i++) cout << " ";
  cout << exporttype() << "\n";
  for (int i = 0; i < indent; i++) cout << " ";
  cout << "{\n";
  
  for (list<symbolborderwidget> *sbw = widgets; sbw != NULL; sbw = sbw->next)
    sbw->elem->export(indent + 2);

  for (int i = 0; i < indent; i++) cout << " ";
  cout << "};\n";
}

void boxwidget::passtochildren(void (widget::*fn)(void *), void *p)
{
  for (list <symbolborderwidget> *l = widgets; l != NULL; l = l->next)
    (l->elem->*fn)(p);
}

boxwidget::boxwidget(attribute *attribs, displaybox *boxs, GtkBox *box, GtkWidget *(*new_separator_func)()) : symbolwidget(attribs, boxs)
{
  list<symbolwidget> *contents = (list<symbolwidget> *) NULL;
  int tnameval;
  list<int> tname(&tnameval, (list<int> *) NULL);
  symbol *s;
  
  for (list<symbolconstructor> *l = boxs->constructors; l != NULL; l = l->next)
  {
    tnameval = l->elem->getname();
    s = boxs->getlocalsymbol(&tname);
    if (s == NULL) 
    {
      error(TYPE_ERROR, -1, NULL, "Can't find %s to create box %s.", symbtable.tostring(tnameval), boxs->getfullnametext());
      continue;
    }
    contents = new list<symbolwidget>(s->getwidget(), contents);
  }

  // Now that we have the symbolwidgets, make the symbolborderwidgets
  // #### Memory leak
  widgets = (list<symbolborderwidget> *) NULL;
  for (;contents != NULL; contents = contents->next)
    widgets = new list<symbolborderwidget>(new symbolborderwidget(attribs, contents->elem), widgets);

  // Time to create the various widgets
  boxwidget::new_separator_func = new_separator_func;

  wi = GTK_WIDGET(box);
  //  gtk_container_set_border_width(GTK_CONTAINER(wi), 5);
  
  lastseparator = newseparator();
  for (list<symbolborderwidget> *t = widgets; t != NULL; t = t->next)
  {
    gtk_box_pack_start(box, t->elem->GrabGtkWidget(this, lastseparator), t->elem->canexpand(), TRUE, 0);
    lastseparator = newseparator();
  }
}

GtkWidget *boxwidget::newseparator()
{
  GtkWidget *w = new_separator_func();
  gtk_widget_set_usize(w, 7, 7);
  gtk_box_pack_start(GTK_BOX(wi), w, FALSE, TRUE, 0);
  gtk_widget_show(w);

// Start drag and drop
  gtk_drag_dest_set(w, GTK_DEST_DEFAULT_ALL, AllTargets, 5, GDK_ACTION_MOVE);

  gtk_signal_connect(GTK_OBJECT (w), "drag-data-received", GTK_SIGNAL_FUNC(boxwidget::usedropdata), this);
  //  gtk_signal_connect(GTK_OBJECT (w), "drag-motion", GTK_SIGNAL_FUNC(test), this);

  return w;
}

void boxwidget::usedropdata(GtkWidget *widget, GdkDragContext *drag_context, gint x, gint y, GtkSelectionData *data, guint info, guint time, boxwidget *_this)
{
  symbolborderwidget *symb;
  
  switch (info)
  {
    case TARGET_MOVE:
      symb= *(symbolborderwidget **) data->data;
      gtk_drag_finish(drag_context, TRUE, TRUE, time);
      break;
    case TARGET_HBOX:
      symb = new symbolborderwidget(_this->initattribs, displaybox::getempty(BOX_H)->getwidget());
      gtk_drag_finish(drag_context, TRUE, TRUE, time);
      break;
    case TARGET_VBOX:
      symb = new symbolborderwidget(_this->initattribs, displaybox::getempty(BOX_V)->getwidget());
      gtk_drag_finish(drag_context, TRUE, TRUE, time);
      break;
    case TARGET_PROJECT:
      symb = new symbolborderwidget(_this->initattribs, Project.getwidget());
      gtk_drag_finish(drag_context, TRUE, TRUE, time);
      break;
    case TARGET_SYMBOL:
      { 
        char *name = (char *) data->data;
        list <int> *n = symbtable.toname(name);
        if (n == NULL) 
        {
          gtk_drag_finish(drag_context, FALSE, TRUE, time);
          return;
        }
        symbol *s = Project.getsymbol(n);
        if (s == NULL) 
        {
          gtk_drag_finish(drag_context, FALSE, TRUE, time);
          return;
        }
        symb = new symbolborderwidget(_this->initattribs, s->getwidget());
        if (symb == NULL) 
        {
          gtk_drag_finish(drag_context, FALSE, TRUE, time);
          return;
        }
        gtk_drag_finish(drag_context, TRUE, TRUE, time);
      }
      break;
    default:
      error(TYPE_INTERNALERROR, -1, NULL, "Internal error : Bad Target type in drag and drop!");
  }

  _this->insertwidget(symb, widget);
  //  gtk_widget_hide_all(gtk_widget_get_toplevel(widget));
  //  gtk_widget_show_all(gtk_widget_get_toplevel(widget));
}

void boxwidget::insertwidget(symbolborderwidget *symb, GtkWidget *psep)
{
  list<symbolborderwidget> *l;

   // Have we moved?
  int wasprev = 0;
  for (l = widgets; l != NULL; l = l->next)
  {
    if (l->elem->GetAssociatedSeparator() == psep) break;
    wasprev = 0;
    if (l->elem->GetAssociatedSeparator() == symb->GetAssociatedSeparator()) wasprev = 1;
  }
  if (wasprev) 
    if (l != NULL || lastseparator == psep)
    {
      symb->display();
      if (symb->GetAssociatedSeparator() != NULL) gtk_widget_show(symb->GetAssociatedSeparator());
      if (symb->getparent()->GetWindowWidget() != NULL) gtk_widget_show(symb->getparent()->GetWindowWidget());
      return; // Not moved
    }

  // This removes symbol from its parent widget.
  GtkWidget *newsep = newseparator();
  GtkWidget *child = symb->GrabGtkWidget(this, newsep);

  // Now we actually insert
  int pos = 0;
  list<symbolborderwidget> **pl = &widgets;
  for (l = widgets; l != NULL; l = *pl, pos++)
  {
    if (l->elem->GetAssociatedSeparator() == psep)
      break;
    pl = &(l->next);
  }

  if (l == NULL && lastseparator != psep) 
    error(TYPE_INTERNALERROR, -1, NULL, "Internal error : boxwidget::insertwidget called with invalid separator\n");

  *pl = new list<symbolborderwidget>(symb, l);
  
  gtk_box_pack_start(GTK_BOX(this->wi), child, symb->canexpand(), TRUE, 0);
  gtk_box_reorder_child(GTK_BOX(this->wi), newsep, pos * 2 + 0);
  gtk_box_reorder_child(GTK_BOX(this->wi), child, pos * 2 + 1);

  symb->display();
  if (symb->GetAssociatedSeparator() != NULL) gtk_widget_show(symb->GetAssociatedSeparator());
  if (symb->getparent()->GetWindowWidget() != NULL) gtk_widget_show(symb->getparent()->GetWindowWidget());
}

void boxwidget::RemoveChild(symbolborderwidget *sbw)
{
  list<symbolborderwidget> **pw = &widgets;
  for (list<symbolborderwidget> *w = widgets; w != NULL; w = *pw)
  {
    if (w->elem == sbw)
    {
      sbw->RemoveFromContainer(this->wi);
      gtk_widget_hide(sbw->GetAssociatedSeparator());
      gtk_widget_destroy(sbw->GetAssociatedSeparator());
      *pw = w->next;
      return;
    }
    pw = &(w->next);
  }

  error(TYPE_INTERNALERROR, -1, NULL, "Internal error : boxwidget::RemoveChild called on non child\n"); 
}

hboxwidget::hboxwidget(attribute *attribs, displaybox *s) : boxwidget(attribs, s,GTK_BOX(gtk_hbox_new(FALSE, 0)), gtk_vseparator_new)
{
}

vboxwidget::vboxwidget(attribute *attribs, displaybox *s) : boxwidget(attribs, s,GTK_BOX(gtk_vbox_new(FALSE, 0)), gtk_hseparator_new)
{
}

// vboxwidget only

void vboxwidget::_runmode()
{
  for (list <symbolborderwidget> *s = widgets; s != NULL; s = s->next)
    gtk_widget_hide(s->elem->GetAssociatedSeparator());

  gtk_widget_hide(lastseparator);
}

void vboxwidget::_editmode()
{
  for (list <symbolborderwidget> *s = widgets; s != NULL; s = s->next)
    gtk_widget_show(s->elem->GetAssociatedSeparator());

  gtk_widget_show(lastseparator);
}

/*consttextwidget::consttextwidget(attribute *attribs, char *text) : symbolwidget(attribs)
{
  wi = gtk_label_new(text);
  // wi = gtk_entry_new();
  // gtk_entry_set_text(GTK_ENTRY(wi), text);
}
*/

/****************************************************************************

				   VARIABLEWIDGET

 ****************************************************************************/

variablewidget::variablewidget(attribute *attribs, variable *var) : symbolwidget(attribs, var)
{
  oldvalue = ~var->getvalue(); // We will display a change at startup
  oldvaluetime = Project.GetTime();

  base = 16;
  usesigned = 0;
  wi = gtk_entry_new();
  gtk_signal_connect(GTK_OBJECT(wi), "activate", GTK_SIGNAL_FUNC(edited), this);
  gtk_signal_connect(GTK_OBJECT(wi), "focus-out-event", GTK_SIGNAL_FUNC(edited2), this);

  width = 0;
  height = 0;
}

void variablewidget::edited(GtkEntry *w, variablewidget *_this)
{
  edited2(w, (GdkEventFocus *) NULL, _this);

  gtk_container_set_focus_child(GTK_CONTAINER(gtk_widget_get_toplevel(GTK_WIDGET(w))), NULL);
  gtk_widget_set_state(GTK_WIDGET(w), GTK_STATE_INSENSITIVE);
  gtk_widget_set_sensitive(GTK_WIDGET(w), TRUE);
}

void variablewidget::edited2(GtkEntry *w, GdkEventFocus *event, variablewidget *_this)
{
  char *txt = gtk_entry_get_text(w);

 ((variable *) _this->getsymbol())->setvalue(txt, _this->base);

 // #### Beep when invalid  

  Project.RedrawAll();
}

void variablewidget::draw()
{
  variable *var = (variable *) getsymbol();
  
  if (oldvalue != var->getvalue())
  {
    char *txt = usesigned ? var->getsignedstringvalue(base) : var->getstringvalue(base);

    GdkFont *fnt = gtk_widget_get_style(wi)->font;
    int ascent, descent, nwidth;
    gdk_string_extents(fnt, txt, (int *) NULL, (int *) NULL, &nwidth, &ascent, &descent);
    width = MAX(nwidth + 8, width);
    height = MAX(ascent + descent + 8, height);

    gtk_entry_set_text(GTK_ENTRY(wi), "");
    gtk_entry_set_text(GTK_ENTRY(wi), txt);
    gtk_widget_set_usize(wi, width, height);

    oldvalue = var->getvalue();
    oldvaluetime = Project.GetTime();

    gtk_widget_set_style(wi, NormalStyle);
    oldvaluestate = 0;
  }
  else if (oldvaluetime < Project.GetTime() && !oldvaluestate)
  {
    oldvaluestate = 1;
    gtk_widget_set_style(wi, InsensitiveStyle);
  }
}

int variablewidget::exportattribs()
{
  cout << "base = " << base << ", signed = " << usesigned;
  return 1;
}

void variablewidget::_parseattribs(attribute *attribs)
{
  int temp;
  if (attribs->getintattribute("base", &temp) && temp >= 2 && temp <= 36) base = temp; 

  if (attribs->getintattribute("signed", &temp)) usesigned = temp;

  variable *var = (variable *) getsymbol();
  if (attribs->getintattribute("randomize", &temp)) var->setvalue(random(1 << var->getsize()));

  if (attribs->getintattribute("clear", &temp)) var->setvalue(0);

  if (attribs->getintattribute("@@addvalue", &temp)) var->setvalue(var->getvalue() + temp);

  if (attribs->getintattribute("@@addbasepow", &temp))
  {
    int val = 1;
    if (temp < 0) val = -1, temp = -temp;
    for (; temp > 0; temp--) val *= base;
    var->setvalue(var->getvalue() + val);
  }

  height = 0;
  width = 0;
  
  oldvalue = ~var->getvalue(); // We will display a change at startup
  oldvaluetime = Project.GetTime();
  draw();
}

/****************************************************************************

				   FLOATWIDGET

 ****************************************************************************/

realwidget::realwidget(attribute *attribs, realsymbol *r) : symbolwidget(attribs, r)
{
  oldvalue = NAN; // We will display a change at startup
  oldvaluetime = Project.GetTime();

  base = 16;
  usesigned = 0;
  wi = gtk_entry_new();
  gtk_signal_connect(GTK_OBJECT(wi), "activate", GTK_SIGNAL_FUNC(edited), this);
  gtk_signal_connect(GTK_OBJECT(wi), "focus-out-event", GTK_SIGNAL_FUNC(edited2), this);

  width = 0;
  height = 0;
}

void realwidget::edited(GtkEntry *w, variablewidget *_this)
{
  edited2(w, (GdkEventFocus *) NULL, _this);

  gtk_container_set_focus_child(GTK_CONTAINER(gtk_widget_get_toplevel(GTK_WIDGET(w))), NULL);
  gtk_widget_set_state(GTK_WIDGET(w), GTK_STATE_INSENSITIVE);
  gtk_widget_set_sensitive(GTK_WIDGET(w), TRUE);
}

void realwidget::edited2(GtkEntry *w, GdkEventFocus *event, variablewidget *_this)
{
  char *txt = gtk_entry_get_text(w);

 ((realsymbol *) _this->getsymbol())->setvalue(atof(txt));

 // #### Beep when invalid  

  Project.RedrawAll();
}

void realwidget::draw()
{
  realsymbol *r = (realsymbol *) getsymbol();

  if (r->getvalue() != oldvalue)
  {
    char txt[40];
    snprintf(txt, 40, "%10.10g", r->getvalue());

    GdkFont *fnt = gtk_widget_get_style(wi)->font;
    int ascent, descent, nwidth;
    gdk_string_extents(fnt, txt, (int *) NULL, (int *) NULL, &nwidth, &ascent, &descent);
    width = MAX(nwidth + 6, width);
    height = MAX(ascent + descent + 6, height);

    gtk_entry_set_text(GTK_ENTRY(wi), txt);
    gtk_widget_set_usize(wi, width, height);

    oldvalue = r->getvalue();
    oldvaluetime = Project.GetTime();

    oldvaluestate = 0;
    gtk_widget_set_style(wi, NormalStyle);
  }
  else if (oldvaluetime < Project.GetTime() && !oldvaluestate)
  {
    oldvaluestate = 1;
    gtk_widget_set_style(wi, InsensitiveStyle);
  }
}

/****************************************************************************

				   PLUGWIDGET

 ****************************************************************************/

plugwidget::plugwidget(attribute *attribs, plug *p) : vboxwidget(attribs, displaybox::getempty(BOX_V))
{
  plugtype *type = p->gettype();
  list<symbolname> *names = (list<symbolname> *) NULL;
  if (type != NULL) names = type->GetSymbolNames();
  int i = 0;

  for (list<symbol> *l = p->symbols; l != NULL; l = l->next)
  {
    symbolborderwidget *s = new symbolborderwidget(attribs, l->elem->getwidget());
    insertwidget(s, lastseparator);
    
    if (names == NULL) s->setname(i);
    else
    {
      s->setname(symbtable.toid(symbtable.tostring(names->elem)));
      names = names->next;
    }
    i++;  
  }

  s = p; // Valid since boxwidget doesn't use its symbol after initialisation. (But horrid)
}

/****************************************************************************

				   DISASSEMBLYWIDGET

 ****************************************************************************/

int disassemblywidget::getfocusaddress()
{
  return GTK_CLIST(listwid)->focus_row;
}

void disassemblywidget::_parseattribs(attribute *attribs)
{
  int dummy;
  if (attribs->getintattribute("@@ToCursor", &dummy))
  {
    ((disassembly * )getsymbol())->ToCursor(getfocusaddress());
    Project.RunMode = RUN_GO;
    setidleidle();
  }

  if (attribs->getintattribute("@@SetBreakPoint", &dummy))
  {
    ((disassembly * )getsymbol())->ToggleBreak(getfocusaddress());
  }

  if (attribs->getintattribute("@@ClearBreakPoint", &dummy))
  {
    ((disassembly * )getsymbol())->ClearBreakpoints();
  }

  if (attribs->getintattribute("@@SetIP", &dummy))
  {
    ((disassembly * )getsymbol())->SetIP(getfocusaddress());
  }

  if (attribs->getintattribute("@@addvalue", &dummy))
  {
    minheight = MAX(0, dummy * 20 + minheight);
    gtk_widget_set_usize(GTK_WIDGET(listwid), -1, minheight);
  }

  if (attribs->getintattribute("height", &dummy))
  {
    minheight = MAX(0, dummy);
    gtk_widget_set_usize(GTK_WIDGET(listwid), -1, minheight);
  }
}

int disassemblywidget::exportattribs()
{
  cout << "height = " << listwid->allocation.height;
  return 1;
}

disassemblywidget::~disassemblywidget()
{
  delete BreakState;
}

disassemblywidget::disassemblywidget(attribute *attribs, disassembly *d) : symbolwidget(attribs, d)
{
  OldIP = -1;
  minheight = 100;
  d->GetMemRange(&saddr, &eaddr);
  BreakState = new char[eaddr - saddr];

  char *Title[] = { " ", "Addr", "Hex", "Disassembly", "Source"};
  wi = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(wi), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  listwid = gtk_clist_new_with_titles(5, Title);
  gtk_container_add(GTK_CONTAINER(wi), listwid);
  gtk_widget_show(listwid);

  //  char format[] = "%0?X";
  //  format[2] = log2(eaddr) / 4 + 1 + '0';
    
  int row = 0;
  for (int i = saddr; i < eaddr; i += d->GetOpCodeLength(i), row++)
  {
    char *contents[5];
    //    char buff[sizeof(int) * 2 + 1];
    //    snprintf(buff, sizeof(int) * 2 + 1, format, i);
    BreakState[i - saddr] = d->IsBreak(i);
    contents[0] = BreakState[i - saddr] ? "=>" : "  ";
    contents[1] = d->GetRealAddr(i); //buff;
    contents[2] = d->GetOpCode(i);
    contents[3] = d->GetLine(i);
    contents[4] = d->GetSourceCode(i);
    gtk_clist_append(GTK_CLIST(listwid), contents);
    gtk_clist_set_row_data(GTK_CLIST(listwid), row, (gpointer) i);
    gtk_clist_set_row_style(GTK_CLIST(listwid), row, BreakState[i - saddr] ? InsensitiveStyle : NormalStyle);
  }
  
  gtk_clist_columns_autosize(GTK_CLIST(listwid));
  gtk_widget_set_usize(GTK_WIDGET(listwid), -1, minheight);
  selectcurline();
  jumptoline();
  gtk_signal_connect(GTK_OBJECT(listwid), "select_row", GTK_SIGNAL_FUNC(disassemblywidget::selectrow), this);
}

void disassemblywidget::selectrow(GtkCList *clist, gint row, gint column, GdkEventButton *event, disassemblywidget *_this)
{
  disassembly *dis = (disassembly *) _this->getsymbol();

  //  if (event) cout << event->type << " " << event->button << "\n";
  if (event != NULL && event->button == 1)
  {
    if (event->type == GDK_2BUTTON_PRESS)
    {    
      dis->ToggleBreak((int) gtk_clist_get_row_data(clist, row));
      Project.RedrawAll();
    }

    if (gtk_clist_row_is_visible(GTK_CLIST(_this->listwid), row) != GTK_VISIBILITY_FULL)
      _this->jumptoline(row);
  }

  if (row != ((disassembly *) _this->getsymbol())->GetIP() - _this->saddr) _this->selectcurline();
}

int disassemblywidget::selectcurline()
{
  int curline = ((disassembly *) getsymbol())->GetIP() - saddr;
  gtk_clist_select_row(GTK_CLIST(listwid), curline, 0);

  return curline;
}

void disassemblywidget::jumptoline(int row)
{
  if (row < 0) row = ((disassembly *) getsymbol())->GetIP() - saddr;
  {
    gtk_clist_moveto(GTK_CLIST(listwid), row, 0, .1, 0);
  }
}

void disassemblywidget::draw()
{
  gtk_clist_freeze(GTK_CLIST(listwid));
  int row = selectcurline();
  if (((disassembly *) getsymbol())->GetIP() != OldIP)
  {
    OldIP = ((disassembly *) getsymbol())->GetIP();
    GTK_CLIST(listwid)->focus_row = OldIP - saddr;
  }
  if (prevtime != Project.GetTime() && gtk_clist_row_is_visible(GTK_CLIST(listwid), row) != GTK_VISIBILITY_FULL)
    jumptoline();

  prevtime = Project.GetTime();  

  row = 0;
  disassembly *dis = (disassembly *) getsymbol();
  for (int i = saddr; i < eaddr; i += dis->GetOpCodeLength(i), row ++)
  {
    int isbreak = dis->IsBreak(i);
    if (isbreak == BreakState[i - saddr]) continue;

    BreakState[i - saddr] = isbreak;
    gtk_clist_set_text(GTK_CLIST(listwid), row, 0, isbreak ? "=>" : "  ");
    gtk_clist_set_row_style(GTK_CLIST(listwid), row, isbreak ? InsensitiveStyle : NormalStyle);
  }
  gtk_clist_thaw(GTK_CLIST(listwid));
}

/****************************************************************************

				   OSCILLOSCOPEWIDGET

 ****************************************************************************/

void oscillowidget::drawevent(GtkWidget *widget, GdkEventExpose *event, oscillowidget *_this)
{
  Oscillo *osc = (Oscillo *) _this->getsymbol();

  for (list<pin_OscilloInput> *p = osc->Inputs; p != NULL; p=p->next)
  {
    int px = 0;
    int py = (int) ((10 - p->elem->buffer[0]) * _this->height / (20));

    for (int i = 1; i < osc->memsize; i++)
    {
      int cx = i * _this->width / osc->memsize;
      int cy = (int) ((10 - p->elem->buffer[i]) * _this->height / (20));
      if (i != osc->index) 
        gdk_draw_line(widget->window, widget->style->fg_gc[widget->state], px, py, cx, cy);
      else
      {
        gdk_draw_point(widget->window, widget->style->fg_gc[widget->state], px, py);
        gdk_draw_point(widget->window, widget->style->fg_gc[widget->state], cx, cy);
      }
      px = cx; py = cy;
      //      DEBUG(cerr << p->elem->buffer[i] << " " << px << " " << py << " " << p->elem->max << " " << p->elem->min << "\n";)
    }
  }
}

void oscillowidget::configureevent(GtkWidget *widget, GdkEventConfigure *event, oscillowidget *_this)
{
  _this->width = event->width;
  _this->height = event->height;
}

void oscillowidget::draw()
{
  Oscillo *osc = (Oscillo *) getsymbol();

  if (osc->index <= 1) return;

  for (list<pin_OscilloInput> *p = osc->Inputs; p != NULL; p=p->next)
  {
    int i = osc->index - 1;

    int px = (i - 1) * width / osc->memsize;
    int py = (int) ((10 - p->elem->buffer[i - 1]) * height / (20));

    int cx = i * width / osc->memsize;
    int cy = (int) ((10 - p->elem->buffer[i]) * height / (20));

    gdk_draw_line(wi->window, wi->style->fg_gc[wi->state], px, py, cx, cy);
  }
}

oscillowidget::oscillowidget(attribute *attribs, Oscillo *o) : symbolwidget(attribs, o)
{
    width = 100;
    height = 100;

    wi = gtk_drawing_area_new();
    gtk_drawing_area_size(GTK_DRAWING_AREA(wi), width, height); // ####
    gtk_signal_connect(GTK_OBJECT(wi), "expose_event", GTK_SIGNAL_FUNC(oscillowidget::drawevent), this);
    gtk_signal_connect(GTK_OBJECT(wi), "configure_event", GTK_SIGNAL_FUNC(oscillowidget::configureevent), this);
}

/****************************************************************************

				   PROJECTWIDGET

 ****************************************************************************/

void projectwidget::_parseattribs(attribute *attribs)
{
  int dummy;
  if (attribs->getintattribute("@@addvalue", &dummy))
  {
    minheight = MAX(0, dummy * 20 + minheight);
    gtk_widget_set_usize(GTK_WIDGET(treewid), -1, minheight);
  }

  if (attribs->getintattribute("height", &dummy))
  {
    minheight = MAX(0, dummy);
    gtk_widget_set_usize(GTK_WIDGET(treewid), -1, minheight);
  }
}

int projectwidget::exportattribs()
{
  cout << "height = " << treewid->allocation.height;
  return 1;
}

GtkCTreeNode *projectwidget::insertnode(GtkCTreeNode *parent, symbol *s)
{
  char *namestr = symbtable.tostring(s->getname());
  GtkCTreeNode *node = gtk_ctree_insert_node(GTK_CTREE(treewid), parent, NULL, &namestr, 0, NULL, NULL, NULL, NULL, FALSE, FALSE);
  gtk_ctree_node_set_row_data(GTK_CTREE(treewid), node, s);
  gtk_ctree_node_set_selectable(GTK_CTREE(treewid), node, FALSE);

  return node;
}

void projectwidget::expand(GtkCTree *ctree, GList *ctnode, projectwidget *_this)
{
  gtk_ctree_pre_recursive_to_depth(ctree, GTK_CTREE_NODE(ctnode), GTK_CTREE_ROW(ctnode)->level + 1, projectwidget::createchildren, _this);

  gtk_clist_columns_autosize(GTK_CLIST(_this->treewid));
}

void projectwidget::createchildren(GtkCTree *ctree, GtkCTreeNode *node, projectwidget *_this)
{
  _this->createchildren(node);
}

void projectwidget::createchildren(GtkCTreeNode *ctnode)
{
  symbol *s = (symbol *) gtk_ctree_node_get_row_data(GTK_CTREE(treewid), ctnode);

  if (GTK_CTREE_ROW(ctnode)->children != NULL) return;

  for (list<symbol> *l = s->getapparentsymbols(); l != NULL; l = l->next)
    if (displayanonymous || !symbtable.isanonymous(l->elem->getname()))
      insertnode(GTK_CTREE_NODE(ctnode), l->elem);

  gtk_ctree_sort_node(GTK_CTREE(treewid), GTK_CTREE_NODE(ctnode));
}

projectwidget::projectwidget(attribute *attribs, symbol *proj) : symbolwidget(attribs, proj)
{
  //  SelectedRow = (GtkCTreeNode *) NULL;
  displayanonymous = 0;
  minheight = 200;

  wi = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(wi), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
  treewid = gtk_ctree_new(1, 0);
  gtk_container_add(GTK_CONTAINER(wi), treewid);
  gtk_widget_show(treewid);

  //  gtk_clist_set_column_auto_resize(GTK_CLIST(treewid), 0, TRUE);
  gtk_clist_columns_autosize(GTK_CLIST(treewid));

  gtk_signal_connect(GTK_OBJECT(treewid), "tree_expand", GTK_SIGNAL_FUNC(projectwidget::expand), this);

  GtkCTreeNode *proj = insertnode(NULL, getsymbol());
  createchildren(proj);
  gtk_ctree_expand(GTK_CTREE(treewid), proj);

  gtk_drag_source_set(treewid, GDK_BUTTON1_MASK, &WidgetTarget, 1, (enum GdkDragAction) (GDK_ACTION_MOVE | GDK_ACTION_COPY));
  gtk_signal_connect(GTK_OBJECT(treewid), "drag-data-get", GTK_SIGNAL_FUNC(projectwidget::getdropdata), this);
  gtk_widget_set_usize(GTK_WIDGET(treewid), -1, minheight);
}

void projectwidget::getdropdata(GtkWidget *widget, GdkDragContext *drag_context, GtkSelectionData *data, guint info, guint time, projectwidget *_this)
{
  if (GTK_CLIST(widget)->focus_row == -1) return;

  GtkCTreeNode *sel = gtk_ctree_node_nth(GTK_CTREE(widget), GTK_CLIST(widget)->focus_row);
  symbol *s = (symbol *) gtk_ctree_node_get_row_data(GTK_CTREE(_this->treewid), sel);
  symbolwidget *sw = s->getwidget();
  symbolborderwidget *sbw = new symbolborderwidget(s->getattributes(), sw);

  gtk_selection_data_set(data, 0, sizeof(symbolborderwidget *) * 8, (guchar *) &sbw, sizeof(symbolborderwidget *));
}

/****************************************************************************

				   ARRAYWIDGET

 ****************************************************************************/

arraywidget::~arraywidget()
{
  for (int i = 0; i < end - start; i++)
    delete cells[i];
  delete cells;
  if (cellsw != NULL)
    for (int i = 0; i < (end - start + 1) / 2; i++)
      delete cellsw[i];
    delete cellsw;
}

void arraywidget::passtochildren(void (widget::*fn)(void *), void *p)
{
  if (!wordsize) 
   for (int i = 0; i < end - start; i++)
     (cells[i]->*fn)(p); 
  else
    if (cellsw != NULL)
      for (int i = 0; i < (end - start + 1) / 2; i++)
        (cellsw[i]->*fn)(p); 
}

arraywidget::arraywidget(attribute *attribs, array *a) : symbolwidget(attribs, (symbol *) a) 
{
  a->getmemrange(&start, &end);
  w = 1; 
  h = end - start;
  wi = gtk_table_new(w + 2, h + 1, FALSE);
  table = wi;
  cells = new (symbolwidget *)[h];
  cellsw = NULL;

  wordsize = 0;

  int addr = start;
  for (int i = 0; i < h; i++, addr++)
  {
    list<int> l(&addr);
    symbol *symb = a->getlocalsymbol(&l);
    if (symb != NULL)
      cells[i] = symb->getwidget();
    else
    {
      char *arrname = a->getfullnametext();
      error(TYPE_FATALERROR, -1, NULL, "Can't find symbol %s/%i while creating widget for array %s", arrname, addr, arrname);
    }
  }

  w = 16;
  h = (end - start + w - 1) / w;
  RebuildArray();
}  

void arraywidget::RebuildArray()
{
  if (cellsw == NULL && wordsize)
  {
    int numwords = (end - start + 1) / 2;
    cellsw = new (symbolwidget *)[numwords];

    int addrname;
    int addr = start;
    char buff[] = "w??????????";
    list<int> l(&addrname);
    for (int i = 0; i < numwords; i++, addr+=2)
    {
      sprintf(buff, "w%i", addr);
      addrname = symbtable.toid(buff);

      symbol *symb = getsymbol()->getlocalsymbol(&l);
      if (symb != NULL)
        cellsw[i] = symb->getwidget();
      else
      {
        char *arrname = getsymbol()->getfullnametext();
        error(TYPE_FATALERROR, -1, NULL, "Can't find symbol %s/%s while creating widget for array %s", arrname, buff, arrname);
      }
    }
  }

  gtk_table_resize(GTK_TABLE(table), w + 2, h + 1);
  
  for (int i = 0; i < end - start; i++)
   if (cells[i]->GetWidget()->parent != NULL)
   {
    gtk_widget_ref(cells[i]->GetWidget());
    gtk_container_remove(GTK_CONTAINER(table), cells[i]->GetWidget());
   }

  if (cellsw != NULL)
    for (int i = 0; i < (end - start + 1) / 2; i++)
      if (cellsw[i]->GetWidget()->parent != NULL)
      { 
        gtk_widget_ref(cellsw[i]->GetWidget());
        gtk_container_remove(GTK_CONTAINER(table), cellsw[i]->GetWidget());
      } 
  
  int addr = start;
  int i = 0;
  gtk_container_forall(GTK_CONTAINER(table), (void (*)(GtkWidget *, void *))gtk_widget_destroy, NULL);
  for (int y = 0; y < h; y++)
  {
    char buff[20];
    snprintf(buff, 20, "%Xh", addr); 
    GtkWidget *Lbl = gtk_label_new(buff);
    gtk_widget_show(Lbl);
    gtk_table_attach(GTK_TABLE(table), Lbl, 0, 1, y, y + 1, GTK_EXPAND, GTK_EXPAND, 0, 0);  
    for (int x = 0; x < w; x++, addr += ((int) wordsize) + 1, i++)
    {
      if (addr >= end) break;

      GtkWidget *curcell;
      if (!wordsize) 
        curcell = cells[i]->GetWidget();
      else 
	curcell = cellsw[i]->GetWidget();
      gtk_table_attach(GTK_TABLE(table), curcell, x + 1, x + 2, y, y + 1, GTK_EXPAND, GTK_EXPAND, 0, 0);
      gtk_widget_show(curcell);
    }
  }

  //  refresh();  
}

void arraywidget::_parseattribs(attribute *attribs)
{
  int chcol;

  int dw;
  if ((chcol = attribs->getintattribute("datawidth", &dw)))
  {
    if (dw == 8) wordsize = 0;
    if (dw == 16) wordsize = 1;
    h = (end - start + (wordsize + 1) * (w - 1)) / w / (wordsize + 1);
  }

  if ((chcol |= attribs->getintattribute("columns", &w)))
  {
    if (w < 1) w = 1;
    h = (end - start + (wordsize + 1) * (w - 1)) / w / (wordsize + 1);
  }

  if (chcol) RebuildArray();
}

int arraywidget::exportattribs()
{
  cout << "datawidth = " << (wordsize ? 16 : 8) << ", columns = " << w;
  return 1;
}
